use local_ip_address::local_ip;
use std::{ io::{Bytes, stdin, BufReader, Error, ErrorKind, Read, Write}, net::{IpAddr, Ipv4Addr, SocketAddrV4, TcpStream}, thread::{self, JoinHandle}, time::Duration};
use image::{Rgb, RgbImage};
use ipnet::{Ipv4Net,PrefixLenError};
use log::{trace,debug,warn,error,info};
//use crate::vdma_facade::feedthrough::VdmaHandle;

///RIT owns the subnet 129.21.0.0/16; probing by attempting to open thousands of TCP streams
///simultaneously is unwise.
const UNSAFE_SUBNET:Result<Ipv4Net,PrefixLenError> = Ipv4Net::new(Ipv4Addr::new(129,21,0,0),16);

///This subnet is universally considered the "local" subnet. This results in 256 addresses, and can
///be probed both safely and efficiently, and as long as the world server on the LAN is running it 
///is functionally guaranteeed to be found.
const HOME_SUBNET:Result<Ipv4Net,PrefixLenError> = Ipv4Net::new(Ipv4Addr::new(192,168,0,0),24);

///Current communications takes place on custom port. 
///TODO: RTSP port is 554; prepare for real world situations
const PORT_NUMBER:u16 = 55001;
///Port probe timeout, for 
const PORT_PROBE_TIMEOUT:Duration = Duration::new(5,0);
///The standard subnet is a /24
const SUBNET_SIZE:u8 = 24;

///Storage of all world serve connection information
pub struct WorldServerConnection{
    ///Store all open TcpStreams
    port_list:Vec<(TcpStream,BufReader<TcpStream>)>,
    ///Unique image identifier
    image_index:u64,
    //VDMA instance
    //vdma:Option<VdmaHandle>,
}
impl WorldServerConnection{    
    pub fn new() -> Result<Self,Error>{
        let host_ip = local_ip().expect("No real IP address!");
        trace!("Host IP: {:?}",host_ip);
        let mut thread_list:Vec<JoinHandle<Result<TcpStream,Error>>> = Vec::new();
        let mut port_list:Vec<(TcpStream,BufReader<TcpStream>)> = Vec::new();
        if let IpAddr::V4(host_v4_ip) = host_ip{
            //Check current subnet
            let net = Ipv4Net::new(host_v4_ip,SUBNET_SIZE);
    
            if let Ok(net) = net{
                //RIT owns 129.21.0.0/16; DO NOT SPAM THIS NETWORK
                if UNSAFE_SUBNET.unwrap().contains(&net){
                    //Request world server address from user
                    let address = manual_address_prompt();
                    if address.is_ok(){
                        let real_address = address.unwrap();
                        let reader = BufReader::new(real_address.try_clone().expect("Failed clone"));
                        port_list.push((real_address,reader));
                        //thread_list won't be added to, so the thread for loop won't execute
                    }
                } else if net.contains(&HOME_SUBNET.unwrap()){
                    //Collect all Hosts on the current subnet
                    info!("aggregating all IPs... This may take several minutes...");
                    for net_address in net.hosts(){
                        thread_list.push(
                            thread::spawn(move ||{
                                let ip_port = std::net::SocketAddr::V4(SocketAddrV4::new(net_address, PORT_NUMBER));
                                //Attempt to open a TCP stream with this IP address, exit the thread
                                //returning the open TcpStream, or an error
                                return TcpStream::connect_timeout(&ip_port, PORT_PROBE_TIMEOUT);
                            })
                        );
                    }
                }
            }
        }
    
        //Multithread the port-sniffing, only save sucessful connections
        for thread in thread_list{
            let output = thread.join().unwrap();
            if let Ok(real_output) = output{
                let reader = BufReader::new(real_output.try_clone().expect("Failed clone"));
                port_list.push((real_output,reader));
            }
        }

        return Ok(Self{port_list, image_index:0});//, vdma:None});
    }
    
    
    pub fn get_next_images(&mut self,index:u32){
        for (ref mut stream,ref mut reader) in self.port_list.iter_mut(){
            info!("Stream connected to address: {:?}",stream.peer_addr());
            send_processing_time(stream, index);
            let header_values = read_image_header(stream, reader, &mut self.image_index);
            //if self.vdma.is_none(){
            //    self.vdma = Some(VdmaHandle::new(header_values.0, header_values.1, header_values.2).expect("Vdma init failed!"));
            //}
            read_image(stream, reader, header_values.clone());
            let header_values = read_image_header(stream, reader, &mut self.image_index);
            read_image(stream, reader, header_values.clone());
            self.image_index += 1;
        }
    }
}
///Borrows an iterator over the bytes in a buffered-reader of a TcpStream; reads the next 4 bytes,
///outputs the generated u32
fn bytes_to_u32(iter:&mut Bytes<&mut BufReader<TcpStream>>) -> u32{
    let mut array:[u8;4] = [0;4];
    for i in 0..4{
        array[i] = iter.next().unwrap().unwrap();
    }
    return u32::from_le_bytes(array);
}

fn bytes_to_u64(iter:&mut Bytes<&mut BufReader<TcpStream>>) -> u64{
    let mut array:[u8;8] = [0;8];
    for i in 0..8{
        array[i] = iter.next().unwrap().unwrap();
        debug!("{:?}",array[i]);
    }
    return u64::from_le_bytes(array);
}
///Request the user to input the world server's address into stdin
fn manual_address_prompt() -> Result<TcpStream,Error>{
    let return_val:Result<TcpStream,Error> = Err(Error::new(ErrorKind::AddrNotAvailable,"Invalid Address"));
    while return_val.is_err(){
        let mut user_input:String = String::default();
        match stdin().read_line(&mut user_input){
            Ok(_) => {
                match user_input.trim().parse::<SocketAddrV4>(){
                    Ok(address) => {
                        //Test if the address is accurate
                        let stream = TcpStream::connect_timeout(&std::net::SocketAddr::V4(address), PORT_PROBE_TIMEOUT);
                        if stream.is_ok(){
                            return stream;
                        } else {
                            warn!("Address unavailable!");
                            debug!("{:?}",stream.unwrap_err());
                        }
                    },
                    Err(error) => {
                        error!("Unable to properly parse user input!");
                        debug!("{:?}",error);
                    }
                }
            },
            Err(error) => {
                error!("Unable to read user input!");
                debug!("{:?}",error);
            }
        }
    }
    return return_val;
}

//Used to send the calculated time value to the world server
fn send_processing_time(stream:&mut TcpStream ,sent_value:u32){
        debug!("Begin handshake; sending timestamp float...");
        //Write the f64 value, in bytes, to the stream, and ensure it has been sent
        stream.write(&sent_value.to_le_bytes()).unwrap();
        stream.flush().unwrap();
        debug!("Timestamp sent!");
}

//Used to recieve images from the world server
fn read_image_header(stream:&mut TcpStream, reader:&mut BufReader<TcpStream>, image_index:&mut u64) -> (u32,u32,u32,u64,String){
    //Create an iterator over the stream's bytes
    let mut iter = reader.bytes();

    debug!("Continue handshake, read back image packet length...");
    //The first values sent will always be a u32 containing the length of the message (without the
    //length counted), followed by the image width and height
    let length = bytes_to_u64(&mut iter);
    debug!("Image packet length read as {:?}!",length);

    debug!("Continue handshake, read back image width...");
    let image_width = bytes_to_u32(&mut iter);
    debug!("Image width read as {:?}!",image_width);

    debug!("Continue handshake, read back image height...");
    let image_height = bytes_to_u32(&mut iter);
    debug!("Image height read as {:?}!",image_height);

    //The length of the image is the total number of bytes
    let byte_depth = (length / ((image_width * image_height) as u64)) as u32;
    debug!("Image bitdepth calculated as {:?}!", byte_depth);

    //Confirm to world server the correct number of bytes
    stream.write(&byte_depth.to_le_bytes()).unwrap();
    stream.flush().unwrap();

    debug!("Reading camera source info...");
    let camera_source = bytes_to_u32(&mut iter);
    let camera_name:&str;
    let top_cam_img = "./Top".to_owned() + &image_index.to_string() + ".png";
    let side_cam_img = "./Side".to_owned() + &image_index.to_string() + ".png";
    if camera_source == u32::MIN{
        debug!("Image source defined as \"Top\"!");
        camera_name = &top_cam_img;
    } else if camera_source == u32::MAX {
        debug!("Image source defined as \"Side\"!");
        camera_name = &side_cam_img;
    } else {
        error!("Unknown camera value!!!");
        camera_name = "./unknown.png";
    }

    debug!("Writing {:?} to stream...", camera_source.to_le_bytes());
    stream.write(&camera_source.to_le_bytes()).unwrap();
    stream.flush().unwrap();
    (image_width,image_height,byte_depth,length,camera_name.to_string())
}

fn read_image(stream:&mut TcpStream, reader:&mut BufReader<TcpStream>, header_info:(u32,u32,u32,u64,String)){
    let image_width = header_info.0;
    let image_height = header_info.1;
    let byte_depth = header_info.2;
    let image_length = header_info.3;
    let image_name = header_info.4;

    //Create a storage point for the bytes
    let mut recieved_bytes:Vec<u8> = Vec::new();
    //Create an iterator over the stream's bytes
    let mut iter = reader.bytes();
    //TODO: Remove assumption
    //Always assumes world server will send image
    info!("Start reading...");
    //Recieve the rest of the message
    for _ in 0..image_length{
        recieved_bytes.push(iter.next().unwrap().unwrap_or(0));
    }

    info!("Image successfully recieved");

    //Create a blank pane on which to store the read in pixels
    let mut img = RgbImage::new(image_width,image_height);

    //Create a temporary storage point for u8s when generating pixels
    let mut temp:Vec<u8> = Vec::new();

    //Iterate over bytes
    info!("Building image...");
    for i in 0..recieved_bytes.len(){
        let byte = recieved_bytes.get(i).expect("Index does not exist!");

        temp.push(byte.clone());
        //Each pixel is composed of [byte_depth] u8s
        if ((i as u32)+1) % byte_depth == 0{
            //Coordinate in the pane = i/3 
            let pixel_number:u32 = ((i as u32) /byte_depth).try_into().unwrap();
            //Insert the pixel into the pane
            img.put_pixel(
                //Pixels coming directly from Unity are sent in reference to Unity; that is,
                //starting at the bottom left, and iterating horizontally, then vertically.
                //This line compensates and creates a properly-oriented image
                pixel_number%image_width, (image_height-1)-(pixel_number / image_width), 
                //Extract image from temporary storage, ang generate pixel
                Rgb::<u8>{0:temp.clone().try_into().expect("bad vector length!")}
            );
            //Clear temporary storage
            temp.clear();
        } 
    }
    //Dump image to file
    //ONLY FOR EXPERIMENTATION AND TESTING PURPOSES
    info!("Image built, saving to file...");
    _ = img.save(image_name);
    info!("image saved!");
}
