use std::{collections::VecDeque, fs::{File, OpenOptions}, io::{BufRead, BufReader, Error}, os::unix::fs::OpenOptionsExt, path::Path};
use crate::vdma_facade::O_SYNC;

///Stores the result of mapping a given memory section.
#[derive(Debug,Copy,Clone)]
enum MapResult{
    UNSET, 
    OK,
    FAILED
}

///The value to get back from the filesystem.
///MapAddress and MapSize require the index of the map being initialised
enum MemoryValue{
    MapAddress(usize),
    MapSize(usize),
    Event,
    Name,
    Version,
}

///A virtual memory mapping
///Includes the address, size of the memory block, and whether it was successfully initialised.
#[derive(Debug,Copy,Clone,Default)]
pub struct UioMap{
    address:u32,
    size:u32,
}

impl UioMap {
    ///Getter for the address
    pub fn get_address(&self) -> u32 { self.address }
    ///Getter for the size
    pub fn get_size(&self) -> u32 { self.size }
}

///Information for a given block of virtual memory mappings
#[derive(Debug)]
pub struct UioInfo{
    uio_file:File,
    uio_file_name:String,
    maps:Vec<UioMap>,
    event_count:u64,
    name:String,
    version:String,
    dev_attributes:VecDeque<(String,String)>,
}

impl UioInfo{
    ///Creates a new UioInfo object, if possible
    pub fn new(path:&Path) -> Result<Self,Error> { 
        let uio_file = File::open(path).expect("Unexpected file failure!");
        let uio_file_name:String = path.file_name().expect("Path cannot be parsed!").to_string_lossy().into();
        let mut new_info = UioInfo{
            uio_file,
            uio_file_name,
            maps:Vec::new(),
            event_count:0,
            name:String::default(),
            version:String::default(),
            dev_attributes:VecDeque::new(),
        };
        new_info.get_all_info();
        Ok(new_info)
    }

    ///Pseudo-Display of the UIO device
    pub fn show_device(&self){ 
        println!("{}, name={}, version={}, events={}",
                  self.uio_file_name, self.name, self.version, self.event_count);
    }

    ///Get the UioMap of a particular index from the info, if it exists
    pub fn get_map(&self,index:u32) -> Option<&UioMap>{
        self.maps.get(index as usize)
    }

    ///Pseudo-Display of a particular map within the UIO device
    pub fn show_map(&self, index:usize) { 
        if index > 0 && index < self.maps.len(){
            let map = &self.maps[index];
            println!("Map[{}]: {:?}", index, map );
        }
    }

    ///Pseudo-Display of all maps in the UIO device
    pub fn show_maps(&self) { 
        for i in 0..self.maps.len(){
            self.show_map(i);
        }
    }

    ///Gets the name of the object as found in /sys/class/uio
    pub fn get_name(&self) -> String { self.name.clone() }

    ///Get the file that represents the uio device.
    pub fn get_device(&self) -> File { self.uio_file.try_clone().expect("File cannot be cloned!") }

    ///Get a given memory value from the filesystem.
    ///This should only be used for initialisation.
    fn get_memory_value(&mut self, search_term:MemoryValue){ 
        let uio_path;
        match search_term{
            MemoryValue::MapAddress(map_index) => 
                uio_path = format!("/sys/class/uio/{}/maps/map{}/addr",
                                    self.uio_file_name, map_index),
            MemoryValue::MapSize(map_index) => 
                uio_path = format!("/sys/class/uio/{}/maps/map{}/size",
                                    self.uio_file_name, map_index),
            MemoryValue::Name =>
                uio_path = format!("/sys/class/uio/{}/name",
                                    self.uio_file_name),
            MemoryValue::Event =>
                uio_path = format!("/sys/class/uio/{}/event",
                                    self.uio_file_name),
            MemoryValue::Version => 
                uio_path = format!("/sys/class/uio/{}/version",
                                    self.uio_file_name),
        }
        let possible_file = File::open(uio_path);
        
        match possible_file{
            Ok(file) => {
                //All files contain a single value, to which they are being referenced
                let lines:String = BufReader::new(file).lines().map(|line| line.expect("Bad line!")).collect();
                match search_term{
                    MemoryValue::MapSize(index) => {
                        let filtered_line = lines.strip_prefix("0x").expect("Bad location!");
                        self.maps[index].size = filtered_line.parse().expect("Impossible error");
                    },
                    MemoryValue::MapAddress(index) => {
                        let filtered_line = lines.strip_prefix("0x").expect("Bad location!");
                        self.maps[index].address = filtered_line.parse().expect("Impossible error");
                    },
                    MemoryValue::Event => {
                        self.event_count = lines.parse().expect("This shouldn't happen");
                    },
                    MemoryValue::Name |
                    MemoryValue::Version => {
                        self.name = lines;
                    }
                }
            },
            Err(error) => {
                println!("{:?}",error);
            }
        }
    }

    ///Wrapper for get_memory_value(); runs all iterations for this object
    fn get_all_info(&mut self){ 
        self.get_memory_value(MemoryValue::Name);
        self.get_memory_value(MemoryValue::Version);
        self.get_memory_value(MemoryValue::Event);
        let sys_location = format!("/sys/class/uio/{}/maps",self.uio_file_name);
        for i in 0..Path::new(&sys_location).read_dir().expect("").count(){
            self.get_memory_value(MemoryValue::MapSize(i));
            self.get_memory_value(MemoryValue::MapAddress(i));
        }
    }
}
///Find all devices on the filesystem, and return them as a VecDeque
pub fn find_devices() -> VecDeque<UioInfo>{
     let mut file_options = OpenOptions::new();
     file_options.read(true).write(true).custom_flags(O_SYNC);
     let dev_folder = Path::new("/dev");
     let mut uio_devices:VecDeque<UioInfo> = VecDeque::new();
     if dev_folder.is_dir(){
         for file in dev_folder.read_dir().expect("Invalid path! Check this is being run on Linux"){
             if let Ok(real_file) = file{
                 let raw_filename = real_file.file_name();
                 let filename:String = raw_filename.to_string_lossy().into();
                 if filename.contains("uio"){
                     let uio_device = UioInfo::new(real_file.path().as_path());
                     match uio_device{
                         Ok(real_device) => uio_devices.push_back(real_device),
                         Err(error) => {
                             error!("{:?}",error);
                             panic!();
                         }
                     }
                 }
             }
         }
     }
    uio_devices
}
