use std::{fs::File, io::Error, os::unix::fs::FileExt, thread, time::Duration};
use crate::vdma_facade::common::{find_devices, UioInfo};

///Object for VDMA control
pub struct VdmaHandle{
    /// File pointing to the VDMA control buffer, along with its information
    control_buffer:(File, UioInfo),
    /// Width of the images being parsed; measured in pixels
    width:u32,
    /// Height of the images being parsed; measured in pixels
    height:u32,
    /// Pixel length, measured in bytes
    pixel_length:u32,
    /// Storage of all available framebuffers, along with their information
    framebuffers:Vec<(File,UioInfo)>,
}

/// Gonna be honest, not sure what this thing is.
const OFFSET_PARK_POINTER_REGISTER:u64 = 0x28;

const REGISTER_RESET:u32 = 0;
const NO_MASK_INTERRUPTS:u32 = 0xf;
//Commented = unused
//const TAIL:u32 = 0x34;
//const OFFSET_VERSION:u32 = 0x2c;

//* Register offsets */
//const READ_CONTROL_REGISTER:u64 = 0x00;
//const READ_STATUS_REGISTER:u64 = 0x04;
//const READ_VSIZE:u64 = 0x50;
//const READ_HSIZE:u64 = 0x54;
//const READ_FRAMEDELAY_STRIDE:u64 = 0x58;
//const READ_FRAMEBUFFER_ONE:u64 = 0x5c;
//const READ_FRAMEBUFFER_TWO:u64 = 0x60;
//const READ_FRAMEBUFFER_THREE:u64 = 0x64;
//const READ_FRAMEBUFFER_FOUR:u64 = 0x68;
const WRITE_CONTROL_REGISTER:u64 = 0x30;
const WRITE_STATUS_REGISTER:u64 = 0x34;
const WRITE_IRQ_MASK:u64 = 0x3c;
const WRITE_REGISTER_INDEX:u64 = 0x44;
const WRITE_VSIZE:u64 = 0xa0;
const WRITE_HSIZE:u64 = 0xa4;
const WRITE_FRAMEDELAY_STRIDE:u64 = 0xa8;
const WRITE_FRAMEBUFFER_ONE:u64 = 0xac;
const WRITE_FRAMEBUFFER_TWO:u64 = 0xb0;
const WRITE_FRAMEBUFFER_THREE:u64 = 0xb4;
//const WRITE_FRAMEBUFFER_FOUR:u64 = 0xb8;

///BIT MASKS FOR CONTROL REGISTER
///https://docs.amd.com/r/en-US/pg020_axi_vdma/S2MM_VDMACR-S2MM-VDMA-Control-Register-Offset-30h
///https://docs.amd.com/r/en-US/pg020_axi_vdma/MM2S_VDMACR-MM2S-VDMA-Control-Register-Offset-00h
/// Further reading can be done at the above linked locations, or the pdf in this repository
const CONTROL_REGISTER_START:u32                    = 0x00000001;
const CONTROL_REGISTER_CIRCULAR_PARK:u32            = 0x00000002;
const CONTROL_REGISTER_RESET:u32                    = 0x00000004;
const CONTROL_REGISTER_GENLOCK_ENABLE:u32           = 0x00000008;
//const CONTROL_REGISTER_FRAME_COUNT_ENABLE:u32       = 0x00000010;
const CONTROL_REGISTER_INTERNAL_GENLOCK:u32         = 0x00000080;
//const CONTROL_REGISTER_WRITE_POINTER:u32            = 0x00000f00;
//const CONTROL_REGISTER_FRAMECOUNTER_IRQENABLE:u32   = 0x00001000;
//const CONTROL_REGISTER_DELAYCOUNT_IRQENABLE:u32     = 0x00002000;
//const CONTROL_REGISTER_ERROR_IRQENABLE:u32          = 0x00004000;
//const CONTROL_REGISTER_REPEAT_ENABLE:u32            = 0x00008000;
//const CONTROL_REGISTER_INTERRUPT_FRAME_COUNT:u32    = 0x00ff0000;
//const CONTROL_REGISTER_IRQ_DELAY_COUNT:u32          = 0xff000000;

///BIT MASKS FOR STATUS REGISTER
///https://docs.amd.com/r/en-US/pg020_axi_vdma/S2MM_VDMASR-S2MM-VDMA-Status-Register-Offset-34h
///https://docs.amd.com/r/en-US/pg020_axi_vdma/MM2S_VDMASR-MM2S-VDMA-Status-Register-Offset-04h
/// Further reading can be done at the above linked locations, or the pdf in this repository
const STATUS_REGISTER_HALTED:u32                    = 0x00000001;
//const STATUS_REGISTER_VDMA_INTENAL_ERROR:u32        = 0x00000010;
//const STATUS_REGISTER_VDMA_SLAVE_ERROR:u32          = 0x00000020;
//const STATUS_REGISTER_VDMA_DECODE_ERROR:u32         = 0x00000040;
//const STATUS_REGISTER_START_OF_FRAME_EARLY_ERROR:u32= 0x00000080;
//const STATUS_REGISTER_END_OF_LINE_EARLY_ERROR:u32   = 0x00000100;
//const STATUS_REGISTER_START_OF_FRAME_LATE_ERROR:u32 = 0x00000800;
//const STATUS_REGISTER_FRAME_COUNT_INTERRUPT:u32     = 0x00001000;
//const STATUS_REGISTER_DELAY_COUNT_INTERRUPT:u32     = 0x00002000;
//const STATUS_REGISTER_ERROR_INTERRUPT:u32           = 0x00004000;
//const STATUS_REGISTER_END_OF_LINE_LATE_ERROR:u32    = 0x00008000;
//const STATUS_REGISTER_FRAME_COUNT:u32               = 0x00ff0000;
//const STATUS_REGISTER_DELAY_COUNT:u32               = 0xff000000;

impl VdmaHandle {
    ///Creates a new VDMAHandle
    pub fn new(width:u32,height:u32,pixel_length:u32) -> Result<VdmaHandle,Error>{
        //uio0 is control
        //uio1-3 are framebuffers
        // for each device in /dev/uio*
        let mut framebuffers:Vec<(File,UioInfo)> = Vec::new();
        let mut uio_devices = find_devices();
        //Control device should always be first; dependent upon DeviceTree implementation
        let control_device = uio_devices.pop_front().unwrap();
        let control_buffer = (control_device.get_device(),control_device);

        //Collect all devices
        for device in uio_devices{
            framebuffers.push((device.get_device(),device));
        }

        //Create a new VdmaHandle object 
        let mut new_object = VdmaHandle{ control_buffer,width,height,pixel_length,framebuffers };

        //Start triple buffering.... whatever that means. Kaputa didn't make it very clear,
        //but it seems to be initialisation things
        new_object.vdma_start_triple_buffering();
        
        //Return new initialised object
        Ok(new_object)
    }

    ///Set a flag in the Vdma control buffer
    fn set_vdma_flag(&mut self, offset:u64, value:u32) {
        _ = self.control_buffer.0.write_at(&value.to_le_bytes(), offset);
    }

    ///Get a value from the VDMA control buffer, assuming it exists
    /// TODO: This might need a unique offset, if the status register != control register. 
    /// I was skimming, so I'm not sure how accurate this is.
    fn get_vdma_flag(&self, offset:u64) -> Result<u32,Error> { 
        let mut buf = [0;4];
        let read_length = self.control_buffer.0.read_at(&mut buf, offset);
        match read_length{
            Ok(length) => {
                if length < 4{
                    return Ok(u32::from_le_bytes(buf));
                } else {
                    return self.get_vdma_flag(offset);
                }
            },
            Err(error) => {
                error!("{:?}",error);
                return Err(error);
            },
        }
    }

    ///Check if a flag is set to a certain value
    ///
    ///Checking flags is fairly regular on initialisation, this makes the code further down
    ///easier to read.
    fn flag_is_set(&self, offset:u64, check_val:u32) -> bool{
        match self.get_vdma_flag(offset){
            Ok(return_val) => {
                return_val & check_val == check_val
            },
            Err(_) => { false }
        }
    }

    ///Start triple buffering; aka initialise the buffers
    ///
    ///Still need to poke around and find out what the hell this actually means, but uh...
    ///later me problem. For now, just a translated version of what Kaputa wrote
    fn vdma_start_triple_buffering(&mut self) { 
        self.set_vdma_flag(WRITE_CONTROL_REGISTER, REGISTER_RESET);

        while !self.flag_is_set(WRITE_CONTROL_REGISTER, CONTROL_REGISTER_RESET)
            { thread::sleep(Duration::from_millis(50)); }

        self.set_vdma_flag(WRITE_STATUS_REGISTER, CONTROL_REGISTER_RESET);
        self.set_vdma_flag(WRITE_IRQ_MASK, NO_MASK_INTERRUPTS);

        let interrupt_frame_count = 3;
        //Interrupt frame count is stored in mask 
        //CONTROL_REGISTER_INTERRUPT_FRAME_COUNT: 0x00ff0000
        //The value should, therefore, be shifted left 16
        let start_flag = (interrupt_frame_count << 16) | 
                         //Start VDMA
                         CONTROL_REGISTER_START | 
                         // Enable """mutex""" for r/w register. 
                         //https://docs.amd.com/r/en-US/pg020_axi_vdma/Genlock-Synchronization?tocId=jzZsIGLYleCfhtxUkvGoug
                         CONTROL_REGISTER_GENLOCK_ENABLE |
                         CONTROL_REGISTER_INTERNAL_GENLOCK |
                         //Enable circular buffer rather than parking
                         //https://docs.amd.com/r/en-US/pg020_axi_vdma/Stream-to-Memory-Map-Register-Detail
                         CONTROL_REGISTER_CIRCULAR_PARK;
        self.set_vdma_flag(WRITE_CONTROL_REGISTER,start_flag);

        while  !self.flag_is_set(WRITE_CONTROL_REGISTER, CONTROL_REGISTER_START) || 
                self.flag_is_set(WRITE_STATUS_REGISTER, STATUS_REGISTER_HALTED)
            { thread::sleep(Duration::from_millis(50)); }

        self.set_vdma_flag(WRITE_REGISTER_INDEX, REGISTER_RESET);

        let framebuffers = &self.framebuffers;
        let frame_buf_one = framebuffers.get(0).expect("No framebuffers!");
        let frame_buf_two = framebuffers.get(1).expect("Only one framebuffer!");
        let frame_buf_three = framebuffers.get(2).expect("Not enough framebuffers");

        let frame_buf_one_addr = frame_buf_one.1.get_map(0).expect("Bad map!").get_address();
        let frame_buf_two_addr = frame_buf_two.1.get_map(0).expect("Bad map!").get_address();
        let frame_buf_three_addr = frame_buf_three.1.get_map(0).expect("Bad map!").get_address();

        self.set_vdma_flag(WRITE_FRAMEBUFFER_ONE,   frame_buf_one_addr );
        self.set_vdma_flag(WRITE_FRAMEBUFFER_TWO,   frame_buf_two_addr );
        self.set_vdma_flag(WRITE_FRAMEBUFFER_THREE, frame_buf_three_addr);

        self.set_vdma_flag(OFFSET_PARK_POINTER_REGISTER, REGISTER_RESET);

        self.set_vdma_flag(WRITE_FRAMEDELAY_STRIDE, self.width*self.pixel_length);
        self.set_vdma_flag(WRITE_HSIZE, self.width*self.pixel_length);
        self.set_vdma_flag(WRITE_VSIZE, self.height);
    }
    ///Send image to VDMA
    ///
    ///TODO: Define format being used, be it raw image, or file
    pub fn set_frame() { 
        //amountWritten = vdma_endpoint.write_at(&[u8;arraySize], offset)
        //amountWritten = framebuffer.write(&[u8;arraySize]);
        //From python file:
        //seek back to the beginning of the file (probably won't need to??? Gonna have to check)
        //write to file
        todo!(); 
    }
    ///Recieve image from VDMA
    ///
    ///TODO: Not sure how to get this output to file system yet...
    pub fn get_frame() { 
        //amountRead = vdma_endpoint.read_at(&mut [u8;arraySize], offset)
        //amountRead = framebuffer.read(&mut [u8;arraySize]);
        todo!(); 
    }
}
